%define fat16_free	0x0000
%define fat16_bad	0xfff7
%define fat16_last	0xfff8	; is equal or above this, last sector

struc fat16_boot

.jump_code		resb 3
.os_name		resb 8
.bytes_per_sector	resw 1
.sectors_per_cluster	resb 1
.reserved_sectors	resw 1
.fat_copies		resb 1
.root_entries		resw 1
.sectors		resw 1
.media_descriptor	resb 1
.sectors_per_fat	resw 1
.sectors_per_track	resw 1
.heads			resw 1
.hidden_sectors		resd 1
.long_sectors		resd 1
.disk			resb 1
.reserved_nt		resb 1
.extended_flag		resb 1
.serial_nr		resd 1
.volume_label		resb 11
.fs_type		resb 8
.boot_code		resb 448
.magic_bytes		resw 1

endstruc

struc fat16_dirent

.file_name		resb 8
.extention		resb 3
.attrib			resb 1
.reserved_nt		resb 1
.creation		resb 1
.creation_time		resw 1
.creation_date		resw 1
.access_date		resw 1
.reserved_fat32		resw 1
.write_time		resw 1
.write_date		resw 1
.starting_cluster	resw 1
.file_size		resd 1

endstruc

%if fat16_boot_size != 0x200
%error sizeof(fat16_boot) is not 0x200.
%endif

%if fat16_dirent_size != 0x20
%error sizeof(fat16_dirent) is not 0x20.
%endif

[section .text]

boot_data	equ	0x7c00
		org	0x7c3e

;-- First of all, initialize the data segment

		xor	ax,ax
		mov	ds,ax

;-- Now set the stack pointer to 0x9000:0xffff

		cli
		mov	ax,0x9000
		mov	ss,ax
		mov	sp,0xffff
		sti

;-- Print the booting message

		mov	si,str_welcome
		call	print

;-- "Calculate" FAT area starting sector

		mov	ax,word[boot_data+fat16_boot.reserved_sectors]
		mov	word[fat_sector],ax

;-- Calculate root directory sector

		mov	ax,word[boot_data+fat16_boot.sectors_per_fat]
		mov	dl,byte[boot_data+fat16_boot.fat_copies]
		xor	dh,dh
		mul	dx

		add	ax,word[fat_sector]
		adc	dx,0

		mov	word[root_sector],ax
		mov	word[root_sector+2],dx

;-- Now search through all root entries

		mov	cx,word[boot_data+fat16_boot.root_entries]
.next_sector:
		mov	bx,root_buffer		; this is our one-sector buffer for root directiry entries
		mov	di,1			; read one sector
		call	read_sectors		; read sector, all registers unchanged
		mov	si,0x0010		; 16 directory entries per sector
		xor	di,di			; di+root_buffer points to the current root directiry entry
.next_directory:
		push	cx
		push	si
		push	di

		mov	si,kernel_name		; compare current root dirent with expected kernel filename
		mov	cx,0x0011		; file names are 11 bytes
		call	memicmp

		pop	di
		pop	si
		pop	cx

		or	al,al			; is al 0?
		jz	.kernel_found
		add	di,0x0020		; point to the next direntry

		dec	cx			; was this the last direntry?
		jz	.kernel_not_found	; yes, no kernel found, quit

		dec	si			; this was the last direntry in this sector?
		jz	.next_sector		; yes, load a new sector
		jmp	short .next_directory	; no, just check next direntry in the same sector

.kernel_not_found:
		mov	si,str_kernel_not_found
		call	print
		jmp	short $-2

.kernel_found:
		mov	si,str_kernel_found
		call	print
		jmp	short $-2

		mov	si,di
;-- ds:si now points to the root directiry entry of the kernel, let's load it!



; print
;
; input
; 	ds:si = string to print
;
; output
;
;	ax = undefined
;	si = undefined

print:
		lodsb
		or	al,al
		jz	.quit
		mov	ah,0x0e
		int	0x10
		jmp	short print
.quit:		ret

; long_to_bios
;
; input
;	dx:ax = sector
;
; output
;
;	ax = undefined
;	cx = BIOS track/sector number
;	dh = BIOS head number (0 to 0xff)

long_to_bios:
		push	ax
		push	dx

		mov	ax,word[boot_data+fat16_boot.heads]
		mov	dx,word[boot_data+fat16_boot.sectors_per_track]
		mul	dx
		xchg	ax,cx

;-- cx = heads * sectors per track

		pop	dx
		pop	ax

;-- dx:ax = sector number

		div	cx
		push	ax

;-- push the track number

		xor	ax,ax
		xchg	ax,dx
		mov	cx,word[boot_data+fat16_boot.sectors_per_track]

;-- dx:ax = (head * sectors per track) + sector
;   cx = sectors per track

		div	cx
		xchg	ax,dx
		inc	ax

;-- ax = one-based sector number
;   dx = head

		pop	cx

;-- cx = track number
;   now check if we're on a hard or floppy disk

		cmp	byte[boot_data+fat16_boot.disk],0x80
		jb	.floppy_disk

;-- we're on a hard disk, cx = (track << 6) + sector

		shl	cx,6
		or	cl,al
		jmp	short .hard_disk

.floppy_disk:

;-- we're on a floppy disk, cx = (track << 8) + sector

		mov	ch,cl
		mov	cl,al

.hard_disk:
		mov	dh,dl
		ret

; read_sectors
;
; input
;	dx:ax = sector
;	es:bx = buffer
;	di = number of sectors to read
;
; output
;	all register unchanged

read_sectors:
		cmp	word[sector_buffered],ax
		jnz	.do_read
		cmp	word[sector_buffered+2],dx
		jz	.skip_read
.do_read:
		mov	word[sector_buffered],ax
		mov	word[sector_buffered+2],dx

		push	di
.read_sector:
		push	ax
		push	dx
		push	cx
		push	bx

		push	cx
		call	long_to_bios
		pop	ax
		mov	dl,al

;-- dl = disk
;   cx = track/sector
;   dh = head

.read_again:
		mov	ax,0x0201
		int	0x13
		jc	.read_again

		pop	bx
		pop	cx
		pop	dx
		pop	ax

		inc	ax
		adc	dx,0
		add	bx,0x0200

;-- increment sector number (dx:ax) with 1, and es:bx with 0x0200

		dec	di
		jnz	.read_sector

;-- loop if we still got anything left to read, otherwise return

		pop	di
.skip_read:

		ret

; memicmp
;
; input
;	ds:si = string 0
;	es:di = string 1
;	cx = size
;
; output
;	al = 0 if equal, else nonzero
;	si undefined
;	di undefined
;	cx = 0

memicmp:
		lodsb
		mov	ah,byte[es:di]
		inc	di
		cmp	al,'a'
		jb	.dont_convert_al
		cmp	al,'z'
		ja	.dont_convert_al
		sub	al,('a'-'A')
.dont_convert_al:
		cmp	ah,'a'
		jb	.dont_convert_ah
		cmp	ah,'z'
		ja	.dont_convert_ah
		sub	ah,('a'-'A')
.dont_convert_ah:
		sub	al,ah
		jnz	.quit
		loop	memicmp
.quit:
		ret

[section .data]

str_welcome	db	"Loading Primula...",0x0d,0x0a,0
str_kernel_not_found db	"Kernel not found!",0
str_kernel_found db	"Kernel found!",0

kernel_name	db	"KERNEL  BIN"
fat_sector	dd	0
sector_buffered	dd	0xffffffff

[section .bss]

root_sector	resd	1
root_buffer	resb	0x0200
