#define	PROGRAM_VERSION		"0.6"

/*

	ebfsm - Extended BozOS File System Manager

	Extended BozOS FS file transfer program version 0.6
	Written by Robert stling and Mathias Bandt (aka Bandi)

	http://bozos.sourceforge.net/

	robost86@hotmail.com
	bandi@greifswald-online.de

	This is a program to format/upload files to BozOS disks.

Changelog:

	2002-02-02 Robert stling
	* Added support for the EBFS file system
	* Changed stage 2 boot loader size to 1KB

	2002-01-01 Robert stling:
	* Fixed the open() thing...

	2001-12-07 Robert stling:
	* Added 'bshd' to support hard drives (and hard disk images)

	2001-09-02 Robert stling:
	* Fixed the TFileEntry struct. (Bandi told me to change 'attrib' to dword), I also fixed some numbers
	* Added comments :)
	* Fixed the bug in allocate_sector, now it returns -1 when there is no space left (Bandi told me about this bug)
	* Fixed a bug in create_file, that could have corrupted a disk with all file table entried used
	* Error messages are now printed to stderr, instead of stdout (thanks Bandi)
	* Fixed a bug in delete_file (thanks Bandi for discovering & fixing this bug)
	* Fixed some small stuff in delete_file, not really a bug, just inefficient code
	* Now user can choose if the disk should be bootable or not.
	* Cleaned up some stuff here and there
	* Added the option to write to stdout instead of a device

	2001-09-03 Robert stling

	* Removed the -c option (write to stdout instead of device). I just though it maybe isn't a very good idea, because
	  we also read alot from the device.
	* Added list_files(), and the -l option
	* Added debug option, -b
	* I _think_ it's giving the files the right size now.

	2001-09-04

	* Now it shows a message about too long filenames
	* Added free_space()
	* list_files() shows free space on disk
	* Checks for free disk space before copy
	* If you try to copy a file, and the local file can't be opened, no empty file is created

	2001-10-17 Robert stling

	* Added support for 6 more floppy formats
	* Improved list_files() a bit. Try -l and check yourself :)

	2001-10-18 Robert stling
	
	* Added the -c option, to create a disk image

To do:

	* Make a function to download files from EBFS disks
	* Add support for hard drive partitions

*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "ebfsm.h"

#define LOADER_NAME	"loader.bin"
#define BOOT_NAME	"boot.bin"

int Debug = 0;

void print_bootsector_info(TBootSector *BootSector) {
	printf(	"Tracks:          %hu (0x%04hx)\n"
			"Heads:           %hhu (0x%02hhx)\n"
			"Sectors:         %hhu (0x%02hhx)\n"
			"Size (sectors)   %lu (0x%08lx)\n"
			"Sector table:    %lu (0x%08lx)\n"
			"Bootloader:      %lu (0x%08lx)\n"
			"Bootloader size: %hhu (0x%02hhx)\n"
			"Disk:            0x%02hhx\n",
			BootSector->Tracks,BootSector->Tracks,
			BootSector->Heads,BootSector->Heads,
			BootSector->Sectors,BootSector->Sectors,
			BootSector->Size,BootSector->Size,
			BootSector->SectorTable,BootSector->SectorTable,
			BootSector->BootLoader,BootSector->BootLoader,
			BootSector->BootLoaderSize,BootSector->BootLoaderSize,
			BootSector->Disk
			);
}

dword allocate_sector(int Device, TBootSector* BootSector, dword Prev) {
	dword x=SECTOR_SYSTEM,NewSector=-1;

	// Seek to Sector Table sector
	lseek(Device,(BootSector->BaseSector+BootSector->SectorTable)<<9,SEEK_SET);
	// Wait until we find a free sector (or find end of disk)
	if(Debug) printf("allocate_sector()\nSearching for free sector...\n");
	while(x!=SECTOR_FREE&&NewSector+1<BootSector->Size) {
		NewSector++;
		if(read(Device,&x,sizeof(x))==-1) {
			fprintf(stderr,"Can not read from device.\n");
			return SECTOR_BAD;
		}
	}
	// The sector we are going to allocate, should be the last in a chain
	x=SECTOR_LAST;
	// Go back to the location of the first free sector in the sector table
	lseek(Device,-sizeof(x),SEEK_CUR);
	// And make it SECTOR_LAST
	if(write(Device,&x,sizeof(x))==-1) {
		fprintf(stderr,"Can not write to device.\n");
			return SECTOR_BAD;
	}
	// If this isn't the first sector in a chain, make the one before point to this.
	if(Debug) printf("Making old sector point to new...\n");
	if(Prev!=SECTOR_LAST) {
		// Seek to the sector before in the chain...
		lseek(Device,((BootSector->BaseSector+BootSector->SectorTable)<<9)+(Prev<<2),SEEK_SET);
		// And make it point to this sector (NewSector)
		if(write(Device,&NewSector,sizeof(NewSector))==-1) {
			fprintf(stderr,"Can not write to device.\n");
			return SECTOR_BAD;
		}
	}
	if(Debug) printf("Allocated sector: %lu\n",NewSector);
	return NewSector;
}

dword free_space(int Device, TBootSector* BootSector) {
	dword Size=0,NewSector=-1,x;

	lseek(Device,(BootSector->BaseSector+BootSector->SectorTable)<<9,SEEK_SET);
	// Wait until we find a free sector (or find end of disk)
	if(Debug) printf("free_space()\nCounting then number of free sectors...\n");
	while(NewSector+1<BootSector->Size) {
		NewSector++;
		if(read(Device,&x,sizeof(x))==-1) {
			fprintf(stderr,"Can not read from device.\n");
			return 0xffffffff;
		}
		if(x==SECTOR_FREE) Size++;
	}
	return Size>>1;
}

void print_file_info(int Output, char* Name, dword Size, dword Attrib, word User) {
	char Line[80];
	int i,j;

	strcpy(Line,Name);
	j = 44 - strlen(Line);
	for(i=strlen(Line);i<=44;i++) Line[i]=' ';
	Line[44]=0;
	sprintf(Line,"%s%lu",Line,Size);
	for(i=strlen(Line);i<52;i++) Line[i]=' ';
	Line[52]=0;
	sprintf(Line,"%s%08x",Line,(unsigned int)Attrib);
	for(i=strlen(Line);i<62;i++) Line[i]=' ';
	Line[62]=0;
	sprintf(Line,"%s%04x\n",Line,User);
	write(Output,Line,strlen(Line));
}

void print_file_info_header(void) {
	printf("Name                                        Size    Attrib.   Owner\n");
}

/*
	This function lists the files in the device Device to Output
*/
int list_files(int Device, TBootSector* BootSector, int Output) {
	TFileEntry File;
	char Tmp[0x80];
	int i,j=0;

	// Seek to the file table
	if(Debug) printf("list_files()\nSeeking to file table.\n");
	lseek(Device,(BootSector->BaseSector+BootSector->FileTable)<<9,SEEK_SET);
	print_file_info_header();
	for(i=0;i<BootSector->FileTableSize<<3;i++) {
		if(read(Device,&File,sizeof(File))==-1) return -1;
		if(File.Name[0]!=0) {
			print_file_info(Output,File.Name,File.Size,File.Attrib,File.User);
			j++;
		}
	}
	sprintf(Tmp,"%d file%s, free space: %lu kb\n",j,(j==1)?"":"s",free_space(Device,BootSector));
	write(Output,Tmp,strlen(Tmp));
	return 0;
}

/*
	This function creates an empty file. Returns 0 for OK, < 0 for error. The file handle it returns should be given as a pointer
	in parameter 4.
*/
int create_file(int Device, TBootSector* BootSector, char* Name, TFileEntry* FileDesc, int* FileOffset) {
	TFileEntry File;
	dword FilePos;
	int Nr;

	if(Debug) printf("create_file()\nSeeking to file table.\n");
	// Seek to the file table
	FilePos = lseek(Device,(BootSector->BaseSector+BootSector->FileTable)<<9,SEEK_SET);
	// Find an empty file table entry
	Nr = BootSector->FileTableSize<<3;
	do {
		if(read(Device,&File,sizeof(File))==-1) {
			printf("Can not read from device.\n");
			return -1;
		}
		FilePos+=sizeof(File);
	} while(File.Name[0]!=0&&(--Nr)!=0);
	if(File.Name[0]!=0) {
		if(Debug) printf("Can not find empty file table entry.\n");
		return -1;		// Can not find any empty file table entry :(
	}

	// Seek to the empty file table entry
	FilePos-=sizeof(File);
	// Allocate a sector for our file
	if(Debug) printf("Allocating sector for file...\n");
	if((File.Sector = allocate_sector(Device,BootSector,SECTOR_LAST))==SECTOR_BAD) {
		fprintf(stderr,"Can not allocate sector.\n");
		return -1;
	}
	if(Debug) printf("Sector allocated OK.\n");
	// Set attribs to 0. In future versions (when we've deicided what we want to use the attribs to ;-), we'll give
	// the user the possibility to change this attrib.
	File.Attrib = 0x0333;
	// Yeah, you're going to be able to choose user & group too in future versions ;-)
	File.User = 0;
	File.Group = 0;
	// Fill the reserved field with zeroes
	memset(File.Reserved,0,sizeof(File.Reserved));
	// Fill the name with zeroes
	memset(File.Name,0,sizeof(File.Name));
	if(strlen(Name)>=sizeof(File.Name)) {
		fprintf(stderr,"File name too long.\n");
		return -1;	// We don't want anyone to overflow the file name filed ;)
	}
	strcpy(File.Name,Name);

	// Seek to the file table entry, and write it
	(*FileOffset) = lseek(Device,FilePos,SEEK_SET);
	if(write(Device,&File,sizeof(File))==-1) {
		fprintf(stderr,"Can not write to device.\n");
		return -1;
	}
	if(Debug) printf("File created.\n");

	// Give caller it's file handle
	(*FileDesc) = File;
	return 0;
}

dword file_size(int File) {
	dword Size;

	Size = lseek(File,0,SEEK_END);
	lseek(File,0,SEEK_SET);
	return Size;
}

int copy_file(int Device, TBootSector* BootSector, char* FileName, TFileEntry* FileDesc, int FileOffset) {
	char Buffer[0x200];
	int BufferSize;
	int File;
	dword Sector;

	if(Debug) printf("copy_file()\nOpening local file (%s).\n",FileName);
	// Open the local file we want to copy
	if((File=open(FileName,O_RDONLY))==-1) {
		fprintf(stderr,"Can not open file: %s\n",FileName);
		return -1;
	}
	if(Debug) printf("Checking if there is enogh of space on disk.\n");
	if(file_size(File)>(free_space(Device,BootSector)<<10)) {
		fprintf(stderr,"Not enough space left on device.\n");
		return -1;
	}
	// 'Sector' is now the first sector of file (should be an empty file)
	Sector = FileDesc->Sector;
	// Read (up to) 0x200 bytes
	BufferSize = read(File,Buffer,0x200);
	while(BufferSize>0) {
		if(Debug) printf("Read %d bytes.\n",BufferSize);
		// Seek to the sector we're going to write to
		lseek(Device,(BootSector->BaseSector+Sector)<<9,SEEK_SET);
		// Write to it
		if(Debug) printf("Writing.\n");
		if(write(Device,Buffer,BufferSize)!=BufferSize) {
			fprintf(stderr,"Can not write to device.\n");
			return -1;
		}
		FileDesc->Size += BufferSize;
		// Read more from the local file
		BufferSize = read(File,Buffer,BufferSize);
		// If EOF (or error), break
		if(BufferSize<=0) break;
		// Allocate a new sector (the first sector in the file chain is allocated by create_file)
		Sector = allocate_sector(Device,BootSector,Sector);
		if(Sector==SECTOR_BAD) {
			fprintf(stderr,"Can not allocate sector.\n");
			return -1;
		}
	}
	if(Debug) printf("Updating file entry.\n");
	lseek(Device,FileOffset,SEEK_SET);
	if(write(Device,FileDesc,sizeof(TFileEntry))==-1) {
		fprintf(stderr,"Can not write to device.\n");
	}
	if(Debug) printf("File copy OK.\n");
	return 0;
}

int delete_file(int Device, TBootSector* BootSector, char* FileName) {
	TFileEntry File;
	dword Sector,x;
	int Files;

	if(Debug) printf("delete_file()\nSeacrhing in file table for file to delete.\n");
	// Files is the number of file table entries
	Files=BootSector->FileTableSize<<3;
	// Seek to the file table
	lseek(Device,(BootSector->BaseSector+BootSector->FileTable)<<9,SEEK_SET);
	do {
		// Read file table entry
		if(read(Device,&File,sizeof(File))==-1) {
			printf("Can not read from device.\n");
			return -1;
		}
	// Stop when we find the file we're looking for (or when there are no more file entries)
	} while(strcmp(File.Name,FileName)&&(--Files));
	// If we didn't find the file, return -1
	if(strcmp(File.Name,FileName)) return -1;
	// Seek to the file entry we found

	/*
	lseek(Device,-sizeof(File),SEEK_CUR);
	memset(&File,0,sizeof(File));
	write(Device,&File,sizeof(File));
	Sector=File.Sector;
	*/
	// Hehe, this must be changed to:
	lseek(Device,-sizeof(File),SEEK_CUR);
	Sector=File.Sector;
	memset(&File,0,sizeof(File));
	write(Device,&File,sizeof(File));
	// This code zeroes the file table entry
	if(Debug) printf("File table entry deleted, freeing sectors...\n");

	// Seek to the location in the sector table of the first sector of file
	lseek(Device,((BootSector->BaseSector+BootSector->SectorTable)<<9)+(Sector<<2),SEEK_SET);
	do {
		// Get the next sector of file
		read(Device,&Sector,sizeof(Sector));
		x=SECTOR_FREE;
		lseek(Device,-4,SEEK_CUR);
		// And free it
		write(Device,&x,sizeof(x));
		if(Debug) printf("%lu.\n",Sector);
		lseek(Device,((BootSector->BaseSector+BootSector->SectorTable)<<9)+(Sector<<2),SEEK_SET);
		// Stop when we find SECTOR_LAST
	} while(Sector!=SECTOR_LAST);
	if(Debug) printf("File deleted.\n");
	return 0;
}

int format(int Device, TBootSector* BootSector, int MakeBoot) {
	int	Boot;
	int	BootLoader;
	int BootSize;
	char* Buffer;
	TFileEntry File;
	dword i,x=SECTOR_FREE;

	// Fill the file table entry with zeroes
	memset(&File,0,sizeof(File));

	memset(BootSector->Reserved,0,sizeof(BootSector->Reserved));
	memset(BootSector->Code,0,sizeof(BootSector->Code));

	if(Debug) printf("format()\n");
	if(MakeBoot) {
		if(Debug) printf("Loading " BOOT_NAME "...\n");
		// Try to open boot.bin
		if((Boot=open(BOOT_NAME,O_RDONLY))==-1) {
			fprintf(stderr,"Can not open " BOOT_NAME "\n");
			return -1;
		}
		// Try to read boot code
		if((BootSize=read(Boot,BootSector->Code,sizeof(BootSector->Code)))==-1) {
			close(Boot);
			fprintf(stderr,"Can not read from " BOOT_NAME "\n");
			return 1;
		}
		close(Boot);
		if(Debug) printf(BOOT_NAME " written to disk.\n");
	}
	// Seek to the boot sector
	lseek(Device,BootSector->BaseSector<<9,SEEK_SET);

	if(Debug) printf("Writing boot sector.\n");
	// Write the boot sector
	if(write(Device,BootSector,0x200)==-1) {
		fprintf(stderr,"Can not write boot sector.\n");
		return -1;
	}

	if(MakeBoot) {
		if(Debug) printf("Loading " LOADER_NAME "\n");
		// Try to open bootloader.bin
		if((BootLoader=open(LOADER_NAME,O_RDONLY))==-1) {
			fprintf(stderr,"Can not open " LOADER_NAME "\n");
			return -1;
		}
		// Allocated memory buffer for bootloader.bin
		if((Buffer=(char*)malloc(BootSector->BootLoaderSize<<9))==NULL) {
			close(BootLoader);
			fprintf(stderr,"Can not allocate memory for copying boot loader.\n");
			return -1;
		}
		// Read bootloader.bin
		if((BootSize=read(BootLoader,Buffer,BootSector->BootLoaderSize<<9))==-1) {
			free(Buffer);
			close(BootLoader);
			fprintf(stderr,"Can not read " LOADER_NAME "\n");
			return -1;
		}
		// Seek to the bootloader sector
		lseek(Device,(BootSector->BaseSector+BootSector->BootLoader)<<9,SEEK_SET);
		// And write the bootloader :)
		if(write(Device,Buffer,BootSize)==-1) {
			free(Buffer);		
			close(BootLoader);
			fprintf(stderr,"Can not write boot loader.\n");
			return -1;
		}
		// Free memory buffer, and close bootloader.bin
		free(Buffer);
		close(BootLoader);
		if(Debug) printf(LOADER_NAME " installed.\n");
	}

	// Seek to the sector table
	lseek(Device,(BootSector->BaseSector+BootSector->SectorTable)<<9,SEEK_SET);
	// And fill it work SECTOR_FREEs
	for(i=0;i<BootSector->Size;i++) {
		write(Device,&x,sizeof(x));
	}
	/*
		The following areas are now going to be set to SECTOR_SYSTEM:
		* Boot Sector
		* Boot loader
		* Sector table itself
		* File table
	*/
	x = SECTOR_SYSTEM;
	if(Debug) printf("Allocating boot sector.\n");
	// Boot sector
	lseek(Device,(BootSector->BaseSector+BootSector->SectorTable)<<9,SEEK_SET);
	write(Device,&x,sizeof(x));
	if(Debug) printf("Allocating bootloader sectors.\n");
	// Boot loader
	lseek(Device,((BootSector->BaseSector+BootSector->SectorTable)<<9)+(BootSector->BootLoader<<2),SEEK_SET);
	for(i=0;i<BootSector->BootLoaderSize;i++) write(Device,&x,sizeof(x));
	if(Debug) printf("Allocating sector table sectors.\n");
	// Sector table
	lseek(Device,((BootSector->BaseSector+BootSector->SectorTable)<<9)+(BootSector->SectorTable<<2),SEEK_SET);
	for(i=0;i<(BootSector->Size>>7)+1;i++) write(Device,&x,sizeof(x));
	if(Debug) printf("Allocating file table sectors.\n");
	// File table
	lseek(Device,((BootSector->BaseSector+BootSector->SectorTable)<<9)+(BootSector->FileTable<<2),SEEK_SET);
	for(i=0;i<BootSector->FileTableSize;i++) write(Device,&x,sizeof(x));

	if(Debug) printf("Creating empty file table.\n");
	// Seek to file table
	lseek(Device,(BootSector->BaseSector+BootSector->FileTable)<<9,SEEK_SET);
	// Fill file table with empty entries (the 'File' struct is set to zero in the beginning of this function
	for(i=0;i<(BootSector->FileTableSize<<3);i++) {
		if(write(Device,&File,sizeof(File))==-1) {
			fprintf(stderr,"Can not wrote empty file descriptor.\n");
			return -1;
		}
	}
	if(Debug) printf("Format OK.\n");
	return 0;
}

void Usage(void) {
	fprintf(stderr,
			"./ebfsm [-b] [-c] device [-t disk type] [-z] [-f] [-s] [-d filename]\n"
			"        [-u filename local_filename] [-l]\n"
			"-b: Debug mode\n"
			"-c: Create file (used for creating disk images)\n"
			"-t: Set the type of disk to use\n"
			"-z: Fills the device with zeroes\n"
			"-f: Creates EBFS file system on the disk\n"
			"-s: Creates EBFS file system and makes the disk bootable\n"
			"    (note: there's no need to use -f with this option)\n"
			"-u: Copies the file local_filename to device/filename\n"
			"-d: Deletes the file\n"
			"-l: List files\n\n"
			"The following disk types are currently supported:\n"
			"  1440 - 1440KB 3.5\" (default)\n"
			"  720  - 720KB 3.5\"   1200 - 120KB 5.25\"   360  - 360KB 5.25\"\n"
			"  180  - 180KB 5.25\"  320  - 320KB 5.25\"   160  - 160KB 5.25\"\n"
			"  hd:nr:tracks:heads:sectors\n\n"
			"Notes: (1) Give the parameters in the correct order!\n"
			"       (2) You have to use -z if you use -c\n"
			"       (3) with -t hd:, the nr field is hex, rest is decimal\n"
			);
}

int main(int argc, char** argv) {
	int Device=-1,i;
	int CreateDevice=0;
	int FileOffset,TmpFile;
	char DeviceName[0x40],z=0;
	TBootSector* BootSector;
	TFileEntry File;
	dword a;

	// Set default disk type to 1440KB 3.5"
	BootSector = &bsfd1440;
	BootSector->Magic = 0xaa55;

	printf(	"Extended BozOS File System Manager " PROGRAM_VERSION "\n"
			"Written by Robert Oestling and Mathias Bandt\n\n");

	if(argc<2) {
		Usage();
		return -1;
	}
	// Parse through the commandline
	for(i=1;i<argc;i++) {
		if(argv[i][0]=='-') {
			switch(argv[i][1]) {
				case 'c':
					CreateDevice = 1;
					break;
				case 'b':
					Debug = 1;
					break;
				case 't':
					if(argc<i+2) {
						fprintf(stderr,"No disk type after -t\n");
						Usage();
						return -1;
					}
					i++;
					if(!memcmp(argv[i],"hd:",3)) {
						BootSector = &bshd;
						sscanf(&argv[i][3],"%hhx:%hd:%hhd:%hhd",
							&bshd.Disk,&bshd.Tracks,&bshd.Heads,&bshd.Sectors);
						bshd.Size = bshd.Tracks * bshd.Heads * bshd.Sectors;
						bshd.FileTable = bshd.SectorTable + (bshd.Size >> 7) + 1;

						printf("Using the following disk structure:\n");
						print_bootsector_info(BootSector);
					} else
					if(!strcmp(argv[i],"1440")) {
						BootSector = &bsfd1440;
					} else
					if(!strcmp(argv[i],"720")) {
						BootSector = &bsfd720;
					} else
					if(!strcmp(argv[i],"1200")) {
						BootSector = &bsfd1200;
					} else
					if(!strcmp(argv[i],"360")) {
						BootSector = &bsfd360;
					} else
					if(!strcmp(argv[i],"180")) {
						BootSector = &bsfd180;
					} else
					if(!strcmp(argv[i],"320")) {
						BootSector = &bsfd320;
					} else
					if(!strcmp(argv[i],"160")) {
						BootSector = &bsfd160;
					} else {
						fprintf(stderr,"Invalid floppy type\n");
						return -1;
					}
					BootSector->Magic = 0xaa55;
					break;
				case 'l':
					if(Device==-1) {
						fprintf(stderr,"No device open.\n");
						return -1;
					}
					if(list_files(Device,BootSector,1)==-1) {
						fprintf(stderr,"Error: Can not list files\n");
						return -1;
					}
					break;
				case 'u':
					if(Device==-1) {
						fprintf(stderr,"No device open.\n");
						return -1;
					} else {
						// If we don't have both a local and remote filename, quit
						if(argc<i+3) {
							Usage();
							return -1;
						}
						if((TmpFile=open(argv[i+2],O_RDONLY))==-1) {
							fprintf(stderr,"Can not open local file (%s).\n",argv[i+2]);
							return -1;
						}
						close(TmpFile);						
						// First try to create the file
						if(create_file(Device,BootSector,argv[++i],&File,&FileOffset)==-1) {
							fprintf(stderr,"Can not create %s\n",argv[i-1]);
							return -1;
						}
						// Then copy all data
						if(copy_file(Device,BootSector,argv[++i],&File,FileOffset)==-1) {
							fprintf(stderr,"Can not copy %s\n",argv[i-1]);
							return -1;
						}
					}
					break;
				case 'f':
					if(Device==-1) {
						fprintf(stderr,"No device open.\n");
						return -1;
					} else {
						// Try to format the drive (not bootable)
						if(format(Device,BootSector,0)==-1) {
							fprintf(stderr,"Can not format disk.\n");
							return -1;
						}
					}
					break;
				case 's':
					if(Device==-1) {
						fprintf(stderr,"No device open.\n");
						return -1;
					} else {
						// Try to format the drive (bootable)
						if(format(Device,BootSector,1)==-1) {
							fprintf(stderr,"Can not format disk.\n");
							return -1;
						}
					}
					break;
				case 'd':
					if(Device==-1) {
						fprintf(stderr,"No device open.\n");
						return -1;
					} else {
						// If we don't have any file to delete, quit
						if(argc<i+2) {
							Usage();
							return -1;
						}
						// Try to delete the file
						if(delete_file(Device,BootSector,argv[++i])==-1) {
							fprintf(stderr,"Can not delete file.\n");
							return -1;
						}
					}
					break;
				case 'z':
					if(Device==-1) {
						printf("No device open.\n");
						return -1;
					} else {
						// Seek to the beginning of the disk
						lseek(Device,0,SEEK_SET);
						// Just zero all of it :)
						for(a=0;a<BootSector->Size<<9;a++) write(Device,&z,1);
					}
					break;
			};
		} else {
			strcpy(DeviceName,argv[i]);
			if(CreateDevice) {
				if(Debug) printf("Creating disk image: %s\n",DeviceName);
				if((Device=open(DeviceName,O_WRONLY|O_CREAT,S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH))==-1) {
					fprintf(stderr,"Can not create disk image.\n");
					return -1;
				}
				close(Device);
			}
			if((Device=open(DeviceName,O_RDWR))==-1) {
				fprintf(stderr,"Can not open device.\n");
				return -1;
			}
		}
	}
	close(Device);
	return 0;
}
