; chardev.asm
;
; Robert stling, 2002-05-02
;
; This offers a low-level interface to charachter-based devices.

%define _chardev_asm 0

[bits 32]
[section .text]

%include "chardev.inc"
%include "mem.inc"
%include "llist.inc"
%include "string.inc"


global chardev_add_device
global chardev_get_pointer_from_name
global chardev_get_pointer_from_id

; Note: all drivers should contain some initiation function that does the following:
; (1) Call chardev_add_device to install the device driver.
; (2) Check if any special options are to be used, and call options()
; (3) Call open()


; eax, esi and edi are used for other purposes when it comes to calls to open(), close() etc.
; ecx, edx, ebx and ebp may be used. For return, all registers except eax may be used.
;
; open()
; Called when the driver is registered. The driver should do its own initiation here.
; There may have been a call to options() before to change some pre-intitiation.
;
; close()
; Called when the driver is about to be shut down.
;
; read(ecx = device ID)
; Returns charachter in ecx and error code in edx if edx != 0 (note, edx = 1 means that no data
;  is waiting if you're in non-blocking mode) cl should be assumed to be the charachter, but
;  some devices may want to use the high 24 bits of ecx.
;
; write(ecx = device ID, dl = charachter)
; Returns error code in edx if edx != 0
;
; options(cl = option, ...)
; Driver specific function to change miscellaneous settings. IOCTL with another name :)


; chardev_add_device
;
; Add a charachter based device driver.
;
; ecx = pointer to list of function pointers and driver ID dword
;	ID
;	open()
;	close()
;	read()
;	write()
;	options()
;	name
;
; Returns:
; ecx = pointer to device.
; eax = 0 if OK

chardev_add_device:
		push	esi
		push	edi

		dllist_create chardev_device, chardev_list
		or	eax,eax
		jnz	.quit

		lea	edi,[esi+chardev_device.id]
		xchg	esi,ecx

		movsd			; ID
		movsd			; open()
		movsd			; close()
		movsd			; read()
		movsd			; write()
		movsd			; options()
		movsd			; name
		xor	eax,eax
.quit:
		pop	edi
		pop	esi
		ret




; chardev_get_pointer_from_name
;
; esi = driver name
;
; Returns:
; eax points to driver struct, or eax = 0 if error

chardev_get_pointer_from_name:
		push	ecx
		push	edx
		push	edi

		mov	ecx,dword[chardev_list]
		mov	edx,ecx
.try_next:
		or	ecx,ecx
		jz	.not_found

		mov	edx,ecx
		mov	ecx,dword[edx+chardev_device.next]

		push	esi
		mov	edi,[edx+chardev_device.name]
		call	string_cmp
		pop	esi

		or	eax,eax
		jz	.found
		jmp	short .try_next
.not_found:
		xor	edx,edx
.found:
		xchg	eax,edx
		pop	edi
		pop	edx
		pop	ecx
		ret



; chardev_get_pointer_from_id
;
; Get pointer to device struct from driver ID
;
; ecx = driver ID
;
; returns:
; eax = pointer to struct, 0 if error

chardev_get_pointer_from_id:
		push	edx

		mov	eax,dword[chardev_list]		; eax points to the element to try
		mov	edx,eax				; ..and edx to the previous element
.try_next:
		or	eax,eax				; have we found the last one?
		jz	.not_found			; yes... the searched element was not found.

		mov	edx,eax
		mov	eax,dword[eax+chardev_device.next]	; load the next block

		cmp	dword[edx+chardev_device.id],ecx	; is this the right element?
		jz	.found				; yay!
		jmp	short .try_next			; nope, try again

.found:							; edx now points to the element wanted
		xchg	eax,edx				; ..now eax does
.not_found:						; eax is either 0, or the pointer to the element
		pop	edx
		ret



; chardev_call <register containing driver pointer>,<call type>
;	chardev_open
;	chardev_close
;	chardev_read
;	chardev_write
;	chardev_options


[section .data]

chardev_list dd		0
