; kbd.asm
;
; Robert stling, 2002-02-04
;
; Keyboard driver.

; kbd_special
;
; bit 0		left shift
; bit 1		right shift
; bit 2		left ctrl
; bit 3		right crtl
; bit 4		alt
; bit 5		alt gr

; kbd_status
;
; bit 0		scroll lock
; bit 1		num lock
; bit 2		caps lock

; keymap entry
;
; 0000	byte	scan code (set bit 7 for prefix)
; 0001	byte	normal ascii
; 0002	byte	shifted ascii
; 0003	byte	alt (gr)ed ascii
; 0004	byte	shifted and alt (gr)ed ascii
; 0005	byte	code with caps lock active (and nothing else)
; 0006	byte	code with num lock active (and nothing else)
; 0007	byte	reserved
;
; size should be 8 bytes


%define _kbd_asm 0

[bits 32]
[section .text]


%include "config.inc"

%include "kbd.inc"
%include "mem.inc"
%include "idt.inc"
%include "irq.inc"
%include "chardev.inc"
%include "pipe.inc"
%include "keymaps.inc"

%define kbd_major_version	00
%define kbd_minor_version	01


global kbd_chardev_init
global kbd_chardev_open
global kbd_chardev_close
global kbd_chardev_read
global kbd_chardev_write
global kbd_chardev_options


kbd_chardev_init:
		push	ecx
		push	edx
		push	ebx
		push	esi
		push	edi

		mov	ecx,kbd_device
		call	chardev_add_device
		or	eax,eax
		jnz	.quit

		call	kbd_chardev_open

		xor	eax,eax
.quit:
		pop	edi
		pop	esi
		pop	ebx
		pop	edx
		pop	ecx
		ret


kbd_chardev_open:
		call	kbd_init
		ret


kbd_chardev_close:
		ret



kbd_chardev_read:
		push	eax
		push	esi

		mov	esi,ecx

		mov	ecx,dword[(esi*4) + kbd_inpipe]

		cmp	byte[kbd_blocking],0
		jz	.no_blocking
.wait:
		call	pipe_empty
		or	eax,eax
		jnz	.wait
.no_blocking:

		call	pipe_read
		mov	dword[(esi*4) + kbd_inpipe],ecx
		mov	al,dl
		push	eax
		call	pipe_read
		pop	eax
		mov	dword[(esi*4) + kbd_inpipe],ecx
		mov	ah,dl

		movzx	ecx,ax
		xor	edx,edx
		jmp	short .quit
.error:
		mov	edx,1
.quit:
		pop	esi
		pop	eax
		ret



kbd_chardev_write:
		xor	edx,edx
		ret



; cl = option

; 0x00 - version
; returns:
; cl = minor, ch = major

; 0x01 - check for pending input data
; edx = device ID (kbd nr)
; returns:
; cl = 0 - no data, cl = 1 - data pending

; 0x10 - set blocking mode
; ch = 0 - off, ch = 1 - on
; returns:
; nothing

; 0x11 - activate virtual keyboard
; ch = virtual keyboard to be activated
; returns:
; nothing

; 0x20 - get shift/alt/ctrl and caps/num/scroll status
; returns:
; cl = caps/num/scroll status, ch = shift/alt/ctrl status

; 0x21 - set caps/num/scroll status and update LEDs
; ch = caps/num/scroll status
; returns:
; nothing

; 0x22 - set keymap
; ecx = pointer to keymap
; returns:
; nothing


kbd_chardev_options:
		cmp	cl,0x00				; version request?
		jnz	.not_00				; nope...

		mov	cx,kbd_minor_version + (kbd_major_version << 8)
		ret
.not_00:
		cmp	cl,0x01				; buffer empty check?
		jnz	.not_01				; nope...

		push	eax
		mov	ecx,dword[(edx*4) + kbd_inpipe]
		call	pipe_empty
		xor	al,1
		mov	cl,al
		pop	eax
		ret
.not_01:

		cmp	cl,0x10				; trying to set (non-)blocking mode?
		jnz	.not_10

		mov	byte[kbd_blocking],ch
		ret
.not_10:
		cmp	cl,0x11
		jnz	.not_11

		push	ecx
		movzx	ecx,ch
		mov	dword[kbd_current_device],ecx
		pop	ecx
		ret
.not_11:
		cmp	cl,0x22				; change of keymap?
		jnz	.not_22

		mov	dword[kbd_map],ecx
		ret
.not_22:
		cmp	cl,0x20
		jnz	.not_20

		mov	cl,byte[kbd_status]
		mov	ch,byte[kbd_special]
		ret
.not_20:
		cmp	cl,0x21
		jnz	.not_21

		mov	byte[kbd_status],ah
		call	kbd_update_leds
		ret
.not_21:
		ret



; kbd_init
;
; Initialize keyboard driver.
;
; eax = size of keyboard buffer
;
; returns:
; all registers unchanged

kbd_init:
		push	eax
		push	ebx
		push	ecx
		push	esi
		push	edi

		mov	dword[kbd_current_device],0
		mov	ecx,kbd_max_devices
.next_pipe:
		dec	ecx

		call	pipe_create
		mov	dword[ecx+kbd_inpipe],edi

		or	ecx,ecx
		jnz	.next_pipe

		mov	byte[kbd_special],0	; empty special keys flags
		mov	byte[kbd_status],0	; empty status flags
		mov	byte[kbd_prefix],0

		mov	eax,0x21		; hook interrupt 0x21 (IRQ1)
		mov	ebx,kbd_int
		call	idt_set_interrupt_vector

		mov	cl,1			; enable IRQ1
		call	irq_enable

		call	kbd_update_leds		; set LEDs

		pop	edi
		pop	esi
		pop	ecx
		pop	ebx
		pop	eax

		ret


; kbd_int
;
; The keyboard interrupt handler.

kbd_int:
		push	eax
		push	esi

		in	al,0x60			; get scancode...

		cmp	al,0xfa
		jnz	.no_ack_here

		mov	byte[kbd_ack],al
		jmp	.no_ack
.no_ack_here:

		mov	byte[kbd_ack],0

		cmp	al,0xe0
		jnz	.no_prefix

		mov	byte[kbd_prefix],al
		jmp	.quit_prefix
.no_prefix:

		push	eax
		mov	al,byte[kbd_special]
		and	al,0x03
		cmp	al,0x03
		jnz	.no_reboot

		call	kbd_wait
		mov	al,0xfe
		out	0x64,al
.no_reboot:
		pop	eax


		cmp	al,0x2a			; left shift?
		jz	.do_left_shift
		cmp	al,0x36			; right shift?
		jz	.do_right_shift

		cmp	al,0xaa			; left shift up?
		jz	.do_left_shift_up
		cmp	al,0xb6			; right shift up?
		jz	.do_right_shift_up

		cmp	al,0x1d			; control?
		jz	.do_ctrl
		cmp	al,0x9d			; control up?
		jz	.do_ctrl_up

		cmp	al,0x38			; alt?
		jz	.do_alt
		cmp	al,0xb8			; alt up?
		jz	near .do_alt_up

		cmp	al,0x3a			; caps lock?
		jz	near .do_caps
		cmp	al,0x45			; num lock?
		jz	near .do_num
		cmp	al,0x46			; scroll lock?
		jz	near .do_scroll

		jmp	 .normal_key

.do_left_shift:
		or	byte[kbd_special],0x01
		jmp	.special_done
.do_right_shift:
		or	byte[kbd_special],0x02
		jmp	short .special_done
.do_left_shift_up:
		and	byte[kbd_special],0xfe
		jmp	short .special_done
.do_right_shift_up:
		and	byte[kbd_special],0xfd
		jmp	short .special_done


.do_ctrl:
		cmp	byte[kbd_prefix],0
		jz	.do_left_control
.do_right_control:
		or	byte[kbd_special],0x08
		jmp	short .special_done
.do_left_control:
		or	byte[kbd_special],0x04
		jmp	short .special_done
.do_ctrl_up:
		and	byte[kbd_special],0xfc
		jmp	short .special_done


.do_alt:
		cmp	byte[kbd_prefix],0
		jz	.do_left_alt
.do_right_alt:
		or	byte[kbd_special],0x20
		jmp	short .special_done
.do_left_alt:
		or	byte[kbd_special],0x10
		jmp	short .special_done
.do_alt_up:
		and	byte[kbd_special],0xcf
		jmp	short .special_done

.do_caps:
		xor	byte[kbd_status],0x04
		jmp	short .special_done
.do_num:
		xor	byte[kbd_status],0x02
		jmp	short .special_done
.do_scroll:
		xor	byte[kbd_status],0x01
;		jmp	short .special_done

.special_done:
		jmp	.quit

.normal_key:
		cmp	al,0x80
		jnb	near .quit
		cmp	byte[kbd_prefix],0	; do we have a prefix?
		jz	.dont_add_80		;
		or	al,0x80			; yes, set high bit, to indicate prefix
.dont_add_80:
		mov	ah,al
		xor	al,al

		mov	esi,dword[kbd_map]
.try_key:

		cmp	byte[esi+keymap.scancode],0
		jz	.quit
		cmp	ah,byte[esi+keymap.scancode]
		jnz	.next_key

		test	byte[kbd_special],0x33	; shift or alt?
		jnz	.ascii_shift_or_alt	;
		test	byte[kbd_status],0x04	; caps lock?
		jnz	.ascii_caps		;
		test	byte[kbd_status],0x02	; num lock?
		jnz	.ascii_num		;
.ascii_normal:
		mov	al,byte[esi+keymap.normal]
		jmp	short .key_found
.ascii_caps:
		mov	al,byte[esi+keymap.caps]
		jmp	short .key_found
.ascii_num:
		mov	al,byte[esi+keymap.num]
		jmp	short .key_found
.ascii_shift_or_alt:
		test	byte[kbd_special],0x03	; shift?
		jz	.ascii_alt
		test	byte[kbd_special],0x30	; alt?
		jz	.ascii_shift
						; both!

		mov	al,byte[esi+keymap.shiftalt]
		jmp	short .key_found
.ascii_alt:
		mov	al,byte[esi+keymap.alt]
		jmp	short .key_found
.ascii_shift:
		mov	al,byte[esi+keymap.shift]
		jmp	short .key_found

.next_key:
		add	esi,byte keymap_size
		jmp	short .try_key

.key_found:
		or	al,al			; if key combination doesn't exist, don't generate anything.
		jz	.quit

		push	ecx
		push	edx
		mov	ecx,dword[kbd_current_device]
		mov	ecx,dword[(ecx*4) + kbd_inpipe]
		mov	dl,al
		call	pipe_write
		mov	dl,ah
		call	pipe_write
		pop	edx
		pop	ecx

.quit:
		mov	byte[kbd_prefix],0
		call	kbd_update_leds
.quit_prefix:

		in	al,0x61
		mov	ah,al
		or	al,0x80
		out	0x61,al
		mov	al,ah
		out	0x61,al			; send ack to kbc

.no_ack:
		mov	al,0x20
		out	0x20,al			; send ack to PIC

		pop	esi
		pop	eax
		iret



; kbd_update_leds
;
; Update the keyboard LEDs.
;
; returns:
; all registers unchanged

kbd_update_leds:
		push	eax

		call	kbd_wait
		mov	al,0xed
		out	0x60,al
		call	kbd_wait
		mov	al,byte[kbd_status]
		and	al,0x07
		out	0x60,al

		pop	eax
		ret


kbd_wait:
		push	eax
.wait:
		in	al,0x64
		test	al,0x02
		jnz	.wait

		pop	eax
		ret



[section .data]

kbd_blocking	db	1
kbd_map		dd	kbd_default_map

kbd_device:
		dd	driver_id_kbd
		dd	kbd_chardev_open
		dd	kbd_chardev_close
		dd	kbd_chardev_read
		dd	kbd_chardev_write
		dd	kbd_chardev_options
		dd	kbd_chardev_name

kbd_chardev_name:
		db	"kbd",0

[section .bss]

kbd_status	resb	1			; caps/num/scroll status
kbd_special	resb	1			; shift/crtl/alt status
kbd_prefix	resb	1			; lest prefix, 0 if none
kbd_ack		resb	1			; 0xfa is ack received, 0 else

kbd_inpipe	resd	kbd_max_devices
kbd_current_device resd	1
