; screen.asm
;
; Robert stling, 2002-05-09
;
; Virtual screens. Based on the older code for virtual consoles.


%define _screen_asm 0

%include "config.inc"

%include "screen.inc"
%include "crt.inc"
%include "mem.inc"
%include "chardev.inc"
%include "string.inc"


[bits 32]
[section .text]

; Supported ANSI codes:
;
; ESC[#A		- move cursor up
; ESC[#B		- move cursor down
; ESC[#C		- move cursor right
; ESC[#D		- move cursor left
; ESC[2J		- clear screen
; ESC[#;#H		- move cursor to row #, col #
; ESC[#;#f		- same as above
; ESC[#;..#m		- send all the values to the attribute selector
;
; Attributes:
;
; 05 blink on
; 25 no-blink
; 30 black foreground
; 31 red foreground
; 32 green foreground
; 33 yellow foreground
; 34 blue foreground
; 35 magenta foreground
; 36 cyan foreground
; 37 white foreground
; 39 default foreground
; 40 black background
; 41 red background
; 42 green background
; 43 yellow background
; 44 blue background
; 45 magenta background
; 46 cyan background
; 47 white background
; 49 default background


global screen_chardev_init
global screen_chardev_open
global screen_chardev_close
global screen_chardev_read
global screen_chardev_write
global screen_chardev_options


%define screen_major_version	00
%define screen_minor_version	01


screen_chardev_init:
		push	ecx
		push	edx
		push	ebx
		push	esi
		push	edi

		mov	ecx,screen_device
		call	chardev_add_device
		or	eax,eax
		jnz	.quit

		call	screen_chardev_open

		xor	eax,eax
.quit:
		pop	edi
		pop	esi
		pop	ebx
		pop	edx
		pop	ecx
		ret


screen_chardev_open:
		push	eax
		call	screen_init
		pop	eax
		ret



screen_chardev_close:
		ret



screen_chardev_read:
		xor	edx,edx
		xor	cl,cl
		ret



screen_chardev_write:
		push	eax
		push	ebx

		mov	ebx,ecx
		call	screen_write

		xor	edx,edx

		pop	ebx
		pop	eax
		ret




; cl = option

; 0x00 - version
; returns:
; cl = minor, ch = major

; 0x10 - set blocking mode
; ch = 0 - off, ch = 1 - on
; returns:
; nothing

; 0x11 - activate virtual keyboard
; ch = virtual keyboard to be activated
; returns:
; nothing

; 0x20 - get shift/alt/ctrl and caps/num/scroll status
; returns:
; cl = caps/num/scroll status, ch = shift/alt/ctrl status

; 0x21 - set caps/num/scroll status and update LEDs
; ch = caps/num/scroll status
; returns:
; nothing

; 0x22 - set keymap
; ecx = pointer to keymap
; returns:
; nothing


screen_chardev_options:
		cmp	cl,0x00				; version request?
		jnz	.not_00

		mov	cx,screen_minor_version + (screen_major_version << 8)
		ret
.not_00:
		cmp	cl,0x10				; trying to set (non-)blocking mode?
		jnz	.not_10
		ret					; ignore it.
.not_10:
		cmp	cl,0x11
		jnz	.not_11

		push	eax
		push	ebx

		movzx	ebx,ch
		call	screen_switch		

		pop	ebx
		pop	eax
		ret
.not_11:
		ret




; screen_init
;
; Initialize the screen driver.
;
; Returns:
; eax = 0 if OK, else 1

screen_init:
		push	ecx
		push	edx
		push	ebx
		push	esi


		push	edi
		mov	ecx,(0x10 * screen_default_screens)/4
		mov	edi,screen_command
		xor	eax,eax
		rep	stosd
		pop	edi


		mov	ecx,screen_default_screens
.allocate_next:
		dec	ecx
		mov	eax,screen_size
		call	mem_alloc
		or	esi,esi
		jz	.error

		mov	dword[screen_structs+(ecx*4)],esi

		mov	ebx,esi
		call	crt_create_screen			; allocate screen buffer, initialize the screen struct etc.
		or	eax,eax
		jnz	.error

		call	crt_clear_screen

		or	ecx,ecx
		jnz	.allocate_next

		mov	dword[screen_active],esi
		mov	byte[esi+screen.active],1
		mov	ebx,esi
		call	crt_load_screen
		call	crt_set_cursor_physical

		xor	eax,eax
		jmp	short .quit
.error:
		mov	eax,1
.quit:
		pop	esi
		pop	ebx
		pop	edx
		pop	ecx
		ret



; screen_switch
;
; ebx = console number to switch to
;
; Returns:
; eax = 0 if OK, else 1

screen_switch:
		push	ecx
		push	edx
		push	ebx
		

		cmp	ebx,dword[screen_default_screens]
		jnb	.error

		mov	edx,ebx
		mov	ebx,dword[screen_structs+(ebx*4)]
		cmp	ebx,dword[screen_active]
		jz	.done

		mov	ebx,dword[screen_active]
		call	crt_save_screen
		mov	byte[ebx+screen.active],0

		mov	ebx,dword[screen_structs+(edx*4)]
		mov	dword[screen_active],ebx
		mov	byte[ebx+screen.active],1

		call	crt_load_screen
		call	crt_set_cursor_physical		

.done:
		xor	eax,eax
		jmp	.quit
.error:
		mov	eax,1
.quit:
		pop	ebx
		pop	edx
		pop	ecx
		ret




screen_write:
		push	eax
		push	ecx
		push	edx
		push	ebx

		mov	eax,ebx					; ebx = screen nr
		shl	eax,4
		add	eax,screen_command			; eax points to the buffer of the screen
		cmp	byte[eax],0				; if the "size" field in the buffer is 0, no command is active
		jz	near .no_command_active

		cmp	dl,'A'
		jb	.not_ucase
		cmp	dl,'Z'
		ja	.not_ucase
		jmp	short .char
.not_ucase:
		cmp	dl,'a'
		jb	.not_lcase
		cmp	dl,'z'
		ja	.not_lcase
		jmp	short .char
.not_lcase:
		mov	cl,byte[eax]
		movzx	ecx,cl
		mov	byte[eax+ecx],dl
		inc	byte[eax]
		jmp	.quit
.char:
		mov	cl,byte[eax]
		movzx	ecx,cl
		mov	byte[eax],0
		mov	byte[eax+ecx],dl

		cmp	byte[eax+1],'['
		jnz	near .quit

		cmp	byte[eax+ecx],'H'
		jz	.do_H_f
		cmp	byte[eax+ecx],'f'
		jz	.do_H_f
		jmp	short .not_H_f
.do_H_f:
		call	screen_ansi_cursor_linecol
		jmp	short .quit
.not_H_f:
		cmp	byte[eax+ecx],'J'
		jnz	.not_2_J
		cmp	byte[eax+ecx-1],'2'
		jnz	.not_2_J

		call	screen_ansi_cursor_clearscreen
		jmp	short .quit
.not_2_J:
		cmp	byte[eax+ecx],'m'
		jnz	.not_m

		call	screen_ansi_attrib
		jmp	short .quit
.not_m:
		cmp	byte[eax+ecx],'A'
		jnz	.not_A

		call	screen_ansi_cursor_up
		jmp	short .quit
.not_A:
		cmp	byte[eax+ecx],'B'
		jnz	.not_B

		call	screen_ansi_cursor_down
		jmp	short .quit
.not_B:
		cmp	byte[eax+ecx],'C'
		jnz	.not_C

		call	screen_ansi_cursor_right
		jmp	short .quit
.not_C:
		cmp	byte[eax+ecx],'D'
		jnz	.not_D

		call	screen_ansi_cursor_left
		jmp	short .quit
.not_D:
		jmp	short .quit

.no_command_active:
		cmp	dl,27
		jnz	.no_command

		mov	byte[eax],1
		jmp	short .quit
.no_command:
		mov	ebx,dword[screen_structs+(ebx*4)]
		call	crt_put_char
.quit:
		pop	ebx
		pop	edx
		pop	ecx
		pop	eax
		ret


screen_ansi_cursor_clearscreen:
		push	ebx
		mov	ebx,dword[screen_structs+(ebx*4)]
		call	crt_clear_screen
		pop	ebx
		ret
		

screen_ansi_cursor_linecol:
		push	eax
		push	ebx
		push	esi

		lea	esi,[eax+2]
		call	string_dec_to_int		; convert the first number to an integer, stop at the semicolon.
		mov	ch,bl				; y = row = first nr
		inc	esi				; lets take a jump over the semicolon ;)
		call	string_dec_to_int
		mov	cl,bl				; x = col = second number

		pop	esi
		pop	ebx
		pop	eax

		push	ebx
		mov	ebx,dword[screen_structs+(ebx*4)]
		call	crt_set_cursor
		pop	ebx

		pop	esi
		pop	ebx
		pop	eax
		ret



screen_ansi_cursor_up:
		push	eax
		push	ecx
		push	ebx
		push	esi

		push	ebx
		lea	esi,[eax+2]
		call	string_dec_to_int
		mov	ecx,ebx
		and	ecx,0xff
		pop	ebx
		mov	ebx,dword[screen_structs+(ebx*4)]

.again:
		call	crt_move_cursor_up
		loop	.again

		pop	esi
		pop	ebx
		pop	ecx
		pop	eax
		ret



screen_ansi_cursor_down:
		push	eax
		push	ecx
		push	ebx
		push	esi

		push	ebx
		lea	esi,[eax+2]
		call	string_dec_to_int
		mov	ecx,ebx
		and	ecx,0xff
		pop	ebx
		mov	ebx,dword[screen_structs+(ebx*4)]

.again:
		call	crt_move_cursor_down
		loop	.again

		pop	esi
		pop	ebx
		pop	ecx
		pop	eax
		ret



screen_ansi_cursor_right:
		push	eax
		push	ecx
		push	ebx
		push	esi

		push	ebx
		lea	esi,[eax+2]
		call	string_dec_to_int
		mov	ecx,ebx
		and	ecx,0xff
		pop	ebx
		mov	ebx,dword[screen_structs+(ebx*4)]

.again:
		call	crt_move_cursor_forward
		loop	.again

		pop	esi
		pop	ebx
		pop	ecx
		pop	eax
		ret



screen_ansi_cursor_left:
		push	eax
		push	ecx
		push	ebx
		push	esi

		push	ebx
		lea	esi,[eax+2]
		call	string_dec_to_int
		mov	ecx,ebx
		and	ecx,0xff
		pop	ebx
		mov	ebx,dword[screen_structs+(ebx*4)]

.again:
		call	crt_move_cursor_back
		loop	.again

		pop	esi
		pop	ebx
		pop	ecx
		pop	eax
		ret



screen_ansi_attrib:
		push	eax
		push	ebx
		push	esi
		push	edi

		mov	edi,dword[screen_structs+(ebx*4)]

		lea	esi,[eax+1]
.next:
		inc	esi
		call	string_dec_to_int

		mov	al,byte[edi+screen.attrib]

		cmp	bl,0
		jnz	.not_0
		mov	al,0x07
		jmp	.done
.not_0:
		cmp	bl,5
		jnz	.not_5
		or	al,0x80
		jmp	.done
.not_5:
		cmp	bl,25
		jnz	.not_25
		and	al,0x7f
		jmp	.done
.not_25:
		cmp	bl,30
		jnz	.not_30
		and	al,0xf0
		jmp	.done
.not_30:
		cmp	bl,31
		jnz	.not_31
		and	al,0xf0
		or	al,0x04
		jmp	.done
.not_31:
		cmp	bl,32
		jnz	.not_32
		and	al,0xf0
		or	al,0x02
		jmp	.done
.not_32:
		cmp	bl,33
		jnz	.not_33
		and	al,0xf0
		or	al,0x0e
		jmp	.done
.not_33:
		cmp	bl,34
		jnz	.not_34
		and	al,0xf0
		or	al,0x01
		jmp	.done
.not_34:
		cmp	bl,35
		jnz	.not_35
		and	al,0xf0
		or	al,0x05
		jmp	.done
.not_35:
		cmp	bl,36
		jnz	.not_36
		and	al,0xf0
		or	al,0x03
		jmp	short .done
.not_36:
		cmp	bl,37
		jnz	.not_37
		and	al,0xf0
		or	al,0x0f
		jmp	short .done
.not_37:
		cmp	bl,39
		jnz	.not_39
		and	al,0xf0
		or	al,0x07
		jmp	short .done
.not_39:
		cmp	bl,40
		jnz	.not_40
		and	al,0x0f
		jmp	short .done
.not_40:
		cmp	bl,41
		jnz	.not_41
		and	al,0x0f
		or	al,0x40
		jmp	short .done
.not_41:
		cmp	bl,42
		jnz	.not_42
		and	al,0x0f
		or	al,0x20
		jmp	short .done
.not_42:
		cmp	bl,43
		jnz	.not_43
		and	al,0x0f
		or	al,0x20
		jmp	short .done
.not_43:
		cmp	bl,44
		jnz	.not_44
		and	al,0x0f
		or	al,0x10
		jmp	short .done
.not_44:
		cmp	bl,45
		jnz	.not_45
		and	al,0x0f
		or	al,0x50
		jmp	short .done
.not_45:
		cmp	bl,46
		jnz	.not_46
		and	al,0x0f
		or	al,0x30
		jmp	short .done
.not_46:
		cmp	bl,47
		jnz	.not_47
		and	al,0x0f
		or	al,0x70
		jmp	short .done
.not_47:
		cmp	bl,49
		jnz	.not_49
		and	al,0x0f
		jmp	short .done
.not_49:
.done:
		mov	byte[edi+screen.attrib],al

		cmp	byte[esi],'m'
		jnz	.next

		pop	edi
		pop	esi
		pop	ebx
		pop	eax
		ret






[section .data]

screen_device:
		dd	driver_id_screen
		dd	screen_chardev_open
		dd	screen_chardev_close
		dd	screen_chardev_read
		dd	screen_chardev_write
		dd	screen_chardev_options
		dd	screen_chardev_name

screen_chardev_name:
		db	"screen",0

[section .bss]

screen_active	resd	1
screen_structs	resd	screen_default_screens
screen_command	resb	0x10 * screen_default_screens
