; serial.asm
;
; Robert stling, 2002-04-30 [started porting to asm]
; Robert stling, 2002-05-02 [Most features implemented]
;
; Serial port driver.
;
; Original release 2001-12-14 by Kallu [in C]
; Second release 2001-12-16 by Robert [in C]
;
; 2002-05-09 [Robert stling]:
;  I made a lot of changes, now all input data is stored in a pipe, and 


%define _serial_asm 0

[bits 32]
[section .text]

%include "config.inc"

%include "serial.inc"
%include "mem.inc"
%include "irq.inc"
%include "idt.inc"
%include "chardev.inc"
%include "pipe.inc"

%define serial_major_version	00
%define serial_minor_version	01


global serial_chardev_init
global serial_chardev_open
global serial_chardev_close
global serial_chardev_read
global serial_chardev_write
global serial_chardev_options


serial_chardev_init:
		push	ecx
		push	edx
		push	ebx
		push	esi
		push	edi

		mov	ecx,serial_device
		call	chardev_add_device
		or	eax,eax
		jnz	.quit

		call	serial_chardev_open

		xor	eax,eax
.quit:
		pop	edi
		pop	esi
		pop	ebx
		pop	edx
		pop	ecx
		ret



serial_chardev_open:
		call	serial_init
		ret



serial_chardev_close:
		push	eax

		test	byte[serial_active],1			; is port0 active?
		jz	.not_port0				; nope

		mov	al,0
		call	serial_close
.not_port0:
		test	byte[serial_active],2			; is port1 active?
		jz	.not_port1				; nope

		mov	al,1
		call	serial_close
.not_port1:
		mov	byte[serial_active],0			; just in case :-)

		pop	eax
		ret



; ecx = device ID (0 = serial0, 1 = serial1 ;)
;
; Returns:
; edx = error code (0 if everything is OK)
; ecx = cl = data byte

serial_chardev_read:
		push	eax

		cmp	byte[serial_blocking],0
		jz	.non_blocking
.wait:
		push	ecx
		shl	ecx,2
		add	ecx,serial_inpipe
		mov	ecx,dword[ecx]
		call	pipe_empty
		pop	ecx

		or	eax,eax
		jnz	.wait
.non_blocking:
		mov	ah,cl
		call	serial_read_byte
		movzx	edx,ah					; return error code in edx
		movzx	ecx,al					; returned data in ecx

		pop	eax
		ret



; ecx = device ID, dl = data
;
; Returns: 
; Error code in edx (0 = OK).

serial_chardev_write:
		push	eax

		mov	ah,cl
		mov	al,dl
		call	serial_send_byte

		pop	eax
		ret



; cl = option

; 0x00 - version
; returns:
; cl = minor, ch = major

; 0x01 - check for pending input data
; edx = device ID (port nr)
; returns:
; cl = 0 - no data, cl = 1 - data pending

; 0x10 - set blocking mode
; ch = 0 - off, ch = 1 - on
; returns:
; nothing

; 0x11 - get active ports
; returns:
; cl = bitmap (if bit n is set, port n is active)

serial_chardev_options:
		cmp	cl,0x00				; version request?
		jnz	.not_00				; nope...

		mov	cx,serial_minor_version + (serial_major_version << 8)
		ret
.not_00:
		cmp	cl,0x01				; buffer empty check?
		jnz	.not_01				; nope...

		push	eax
		push	edx

		shl	edx,2
		add	edx,serial_inpipe
		mov	ecx,dword[edx]
		call	pipe_empty
		xor	al,1
		mov	cl,al

		pop	edx
		pop	eax
		ret
.not_01:

		cmp	cl,0x10				; trying to set (non-)blocking mode?
		jnz	.not_10

		mov	byte[serial_blocking],ch
		ret
.not_10:
		cmp	cl,0x11				; active ports requested?
		jnz	.not_11

		mov	cl,byte[serial_active]
		ret
.not_11:

		ret


; serial_init
;
; Initialized all avaiable serial ports (only port0 and port1 supported)
;
; Returns:
; all registers unmodified

serial_init:
		push	eax
		push	esi
		push	edi

		mov	esi,0x400
		mov	edi,serial_base_ports
		movsd						; get the base I/O ports from the BIOS data area

		cmp	word[serial_base_ports+0],0x3f8
		jnz	.port0_not_there

		xor	eax,eax
		call	serial_init_port			; initialize port0

		call	pipe_create
		mov	dword[serial_inpipe],edi		; create the input pipe of port0

		jmp	short .port0_there
.port0_not_there:
		mov	word[serial_base_ports+0],0
.port0_there:

		cmp	word[serial_base_ports+2],0x2f8
		jnz	.port1_not_there

		mov	eax,1
		call	serial_init_port			; initialize port1

		call	pipe_create
		mov	dword[serial_inpipe+4],edi		; create the input pipe of port1

		jmp	short .port1_there
.port1_not_there:
		mov	word[serial_base_ports+2],0
.port1_there:
		pop	edi
		pop	esi
		pop	eax
		ret



; serial_init_port
;
; eax = port number (0 = port0, 1 = port1)
;
; Returns:
; eax = 0 if no error was detected, else eax = 1

serial_init_port:
		push	ecx
		push	edx
		push	ebx

		mov	dx,word[(eax*2)+serial_base_ports]
		or	dx,dx
		jz	near .port_not_there

		cmp	dx,0x3f8
		jnz	.not_3f8

		mov	cl,4					; port 0x3f8 -> IRQ 4
		jmp	short .irq_done
.not_3f8:
		cmp	dx,0x2f8
		jnz	.not_2f8

		mov	cl,3					; port 0x2f8 -> IRQ 3
		jmp	short .irq_done
.not_2f8:
		mov	eax,1
		jmp	.quit
.irq_done:
		push	eax
		push	ecx
		mov	ah,1
		mov	cl,al
		shl	ah,cl
		or	byte[serial_active],ah
		pop	ecx
		pop	eax

		mov	ch,byte[(eax*8)+serial_config_data+serial_config.partity]
		or	ch,byte[(eax*8)+serial_config_data+serial_config.stopbit]
		or	ch,byte[(eax*8)+serial_config_data+serial_config.wordlength]
		mov	bx,word[(eax*8)+serial_config_data+serial_config.baudrate]

		; cl = IRQ
		; ch = config byte for COMLCR port
		; bx = baud rate

		push	eax
		push	ebx

		mov	al,cl
		add	al,0x20					; We want the interrupt nr in eax
		movzx	eax,al

		cmp	al,0x24
		jnz	.not_int24

		mov	ebx,serial_int_24
		jmp	short .int_done
.not_int24:
		cmp	al,0x23
		jnz	.not_int23

		mov	ebx,serial_int_23
		jmp	short .int_done
.not_int23:
		mov	eax,1
		jmp	.quit
.int_done:
		call	idt_set_interrupt_vector

		pop	ebx
		pop	eax


		add	dl,serial_lcr
		in	al,dx
		or	al,0x80
		out	dx,al
		sub	dl,serial_lcr

		;add	dx,serial_port				; No need to add 0 :D
		mov	al,bl					; low byte of baud rate
		out	dx,al

		add	dl,serial_ier
		mov	al,bh					; high byte of baud rate
		out	dx,al
		sub	dl,serial_ier


		add	dl,serial_lcr
		mov	al,ch					; configuration bits
		out	dx,al
		sub	dl,serial_lcr

		add	dl,serial_iir
		mov	al,0xc7
		out	dx,al
		sub	dl,serial_iir

		add	dl,serial_mcr
		mov	al,0x0b
		out	dx,al
		sub	dl,serial_mcr

		add	dl,serial_ier
		mov	al,0x01
		out	dx,al
		sub	dl,serial_ier

		call	irq_enable				; cl is already the IRQ :)

		xor	eax,eax					; return 0

.port_not_there:
.quit:

		pop	ebx
		pop	edx
		pop	ecx
		ret



serial_int_24:
		push	eax
		push	ecx
		push	edx
.again:
		mov	dx,0x3f8+serial_lsr			; check for pending input data
		in	al,dx
		and	al,0x01
		jz	.done					; nothing left to read, return

		mov	dx,0x3f8
		in	al,dx

;		mov	ah,0
;		call	serial_write_byte			; write the byte to the buffer

		mov	dl,al
		mov	ecx,dword[serial_inpipe]
		call	pipe_write				; write to the input pipe of port0

		jmp	short .again
.done:
		mov	al,0x20
		out	0x20,al

		pop	edx
		pop	ecx
		pop	eax

		iret




serial_int_23:
		push	eax
		push	edx
.again:
		mov	dx,0x2f8+serial_lsr			; check for pending input data
		in	al,dx
		and	al,0x01
		jz	.done					; nothing left to read, return

		mov	dx,0x2f8
		in	al,dx

;		mov	ah,1
;		call	serial_write_byte			; write the byte to the buffer

		mov	dl,al
		mov	ecx,dword[serial_inpipe+4]
		call	pipe_write				; write to the input pipe of port1

		jmp	short .again
.done:
		mov	al,0x20
		out	0x20,al

		pop	edx
		pop	eax

		iret



; serial_buffer_empty
;
; Check if the buffer is empty.
;
; al = port nr
;
; Returns:
; eax = 1 if true, eax = 0 if false

%if 0
serial_buffer_empty:
		push	ecx

		movzx	eax,al
		mov	ecx,dword[(eax*4)+serial_inptr]
		cmp	ecx,dword[(eax*4)+serial_outptr]
		jz	.empty

		xor	eax,eax					; return 0 (not empty)
		jmp	short .quit
.empty:
		mov	eax,1					; return 1 (empty)
.quit:
		pop	ecx
		ret
%endif


; serial_write_byte
;
; Writes a byte to the buffer. The buffer must NOT be empty!
;
; ah = port nr
; al = data

%if 0
serial_write_byte:
		push	ecx
		push	edx
		push	ebx

		movzx	edx,ah					; edx = port nr
		mov	ecx,dword[(edx*4)+serial_buffer]	; ecx points to the buffer
		mov	ebx,dword[(edx*4)+serial_inptr]		; ebx = inptr
		mov	byte[ecx+ebx],al			; buffer[inptr] = al
		inc	ebx					; inptr++
		cmp	ebx,serial_buffer_size			; have we reached the end of the buffer?
		jb	.buffer_ok

		xor	ebx,ebx					; yes, go to the beginning
.buffer_ok:
		mov	dword[(edx*4)+serial_inptr],ebx

		pop	ebx
		pop	edx
		pop	ecx

		ret
%endif


; serial_read_byte
;
; Reads a byte from the buffer.
;
; ah = port nr
;
; Returns:
; al = data
; ah = 0 if byte is read, else 1
; rest of eax undefined

serial_read_byte:
		push	ecx
		push	edx
		push	ebx

		movzx	ecx,ah
		shl	ecx,2
		add	ecx,serial_inpipe			; ecx points to the pipe to use
		mov	ebx,ecx
		mov	ecx,dword[ecx]

		call	pipe_read
		mov	ah,al					; move status from al to ah
		mov	al,dl					; move data from dl to al

		mov	dword[ebx],ecx				; make the pipe pointer point right :)

		pop	ebx
		pop	edx
		pop	ecx
		ret



%if 0
		push	ecx
		push	edx
		push	ebx

		movzx	edx,ah					; edx = port nr
		mov	ecx,dword[(edx*4)+serial_buffer]	; ecx points to the buffer
		mov	ebx,dword[(edx*4)+serial_outptr]	; ebx = outptr

		cmp	ebx,dword[(edx*4)+serial_inptr]
		jz	.error

		mov	al,byte[ecx+ebx]			; buffer[outptr] = al
		inc	ebx					; outptr++
		cmp	ebx,serial_buffer_size			; have we reached the end of the buffer?
		jb	.buffer_ok

		xor	ebx,ebx					; yes, go to the beginning
.buffer_ok:
		mov	dword[(edx*4)+serial_outptr],ebx

		mov	ah,0
		jmp	short .ok
.error:
		mov	ah,1
.ok:
		pop	ebx
		pop	edx
		pop	ecx

		ret
%endif


; serial_close
;
; Shuts down a port. Make sure it's open before you close it...
;
; al = port nr
;
; Returns:
; all regusters unmodified

serial_close:
		push	eax
		push	ecx
		push	edx

		movzx	eax,al

		mov	ecx,eax
		shl	ecx,2

		add	ecx,serial_inpipe
		push	ecx
		mov	ecx,dword[ecx]
		call	pipe_close
		pop	ecx
		mov	dword[ecx],0

		push	eax
		push	ecx

		mov	ah,1
		mov	cl,al
		shl	ah,cl
		not	ah
		and	byte[serial_active],ah

		pop	ecx
		pop	eax

		mov	cl,byte[eax+serial_irq]
		call	irq_disable				; disable IRQ for the port
		
		mov	dx,word[(eax*2)+serial_base_ports]	; get base I/O port
		add	dl,serial_mcr
		mov	al,0					; send 0 to the MCR port
		out	dx,al

		pop	edx
		pop	ecx
		pop	eax
		ret



; serial_send_byte
;
; Send a byte :)
;
; ah = port nr
; al = data
;
; Returns:
; all registers unmodified

serial_send_byte:
		push	eax
		push	ecx
		push	edx

		mov	cl,al
		movzx	eax,ah
		mov	dx,word[(eax*2)+serial_base_ports]	; get base I/O port
		add	dl,serial_lsr
.wait:
		in	al,dx
		and	al,0x20
		jz	.wait

		sub	dl,serial_lsr
		mov	al,cl
		out	dx,al

		pop	edx
		pop	ecx
		pop	eax
		ret



[section .data]

serial_irq	db	0x04,0x03
serial_active	db	0x00
serial_blocking	db	0x01

serial_device:

		dd	driver_id_serial
		dd	serial_chardev_open
		dd	serial_chardev_close
		dd	serial_chardev_read
		dd	serial_chardev_write
		dd	serial_chardev_options
		dd	serial_chardev_name

serial_chardev_name:
		db	"serial",0

serial_config_data:

		dw	serial_default_baudrate
		db	serial_default_stopbit
		db	serial_default_wordlength
		db	serial_default_partity
		db	0,0,0

		dw	serial_default_baudrate
		db	serial_default_stopbit
		db	serial_default_wordlength
		db	serial_default_partity
		db	0,0,0


[section .bss]

serial_base_ports	resw	2	; port0 and port1

serial_inpipe		resd	2	; input pipes

;serial_buffer		resd	2	; 2 buffers
;serial_inptr		resd	2	; input pointers for the 2 buffers
;serial_outptr		resd	2	; output pointers for the 2 buffers
